%name PDDL_Parser
%define ERROR log_error
%define ERROR_BODY = 0
%define ERROR_VERBOSE 1
%define LEX next_token
%define LEX_BODY = 0
%define DEBUG 1

%define INHERIT : public PDDL_Base
%define CONSTRUCTOR_PARAM StringTable& t
%define CONSTRUCTOR_INIT : PDDL_Base(t), error_flag(false), \
  stored_n_param(0), current_atom(0)
%define MEMBERS \
public: \
virtual std::ostream& syntax_errors() = 0; \
bool    error_flag; \
private: \
variable_vec        current_param; \
size_t              stored_n_param; \
AtomBase*           current_atom; \
std::list<Effect *> context; \

%header{
#include <stdlib.h>
#include <string.h>
#include <list>
#include "base.h"
%}

%union {
  StringTable::Cell*     sym;
  PDDL_Base::Clause*     clause;
  PDDL_Base::OneOf*      oneof;
  int                    ival;
}

%token TK_OPEN TK_CLOSE TK_OPEN_SQ TK_CLOSE_SQ TK_EQ TK_HYPHEN

%token <sym>  TK_NEW_SYMBOL TK_OBJ_SYMBOL TK_TYPE_SYMBOL TK_PRED_SYMBOL
              TK_FUN_SYMBOL TK_VAR_SYMBOL TK_ACTION_SYMBOL TK_MISC_SYMBOL
              TK_KEYWORD TK_NEW_VAR_SYMBOL
%token <ival> TK_INT

%token KW_REQS KW_CONSTANTS KW_PREDS KW_TYPES KW_DEFINE KW_DOMAIN
       KW_ACTION KW_ACTION KW_ARGS KW_PRE KW_COND KW_EFFECT KW_AND
       KW_OR KW_EXISTS KW_FORALL KW_IMPLY KW_NOT KW_WHEN KW_ONEOF
       KW_PROBLEM KW_FORDOMAIN KW_OBJECTS KW_INIT KW_GOAL KW_NAME 

%type <sym>    arg_symbol action_symbol any_symbol
%type <clause> cnf_clause
%type <oneof>  oneof_effect

%start pddl_declarations

%%

pddl_declarations:
pddl_domain pddl_declarations
| pddl_problem pddl_declarations
| /* empty */
;

pddl_domain:
TK_OPEN KW_DEFINE domain_name domain_elements TK_CLOSE
| TK_OPEN KW_DEFINE domain_name error TK_CLOSE
{
  log_error( "syntax error in domain" );
  yyerrok;
}
;

domain_elements:
domain_requires domain_elements
| domain_types domain_elements
| domain_constants domain_elements
| domain_predicates domain_elements
| domain_structure domain_elements
| /* empty */
;

domain_name:
TK_OPEN KW_DOMAIN any_symbol TK_CLOSE
{
  domain_name = $3->text;
}
;

any_symbol:
TK_NEW_SYMBOL { $$ = $1; }
| TK_OBJ_SYMBOL { $$ = $1; }
| TK_TYPE_SYMBOL { $$ = $1; }
| TK_PRED_SYMBOL { $$ = $1; }
| TK_FUN_SYMBOL { $$ = $1; }
| TK_VAR_SYMBOL { $$ = $1; }
| TK_ACTION_SYMBOL { $$ = $1; }
| TK_MISC_SYMBOL { $$ = $1; }
;

arg_symbol:
TK_OBJ_SYMBOL { $$ = $1; }
| TK_VAR_SYMBOL { $$ = $1; }
;

action_symbol:
TK_NEW_SYMBOL { $$ = $1; }
| TK_ACTION_SYMBOL { $$ = $1; }
;

// requirement declarations

domain_requires:
TK_OPEN KW_REQS require_list TK_CLOSE
| TK_OPEN KW_REQS error TK_CLOSE
{
  log_error( "syntax error in requirements declaration." );
  yyerrok; 
}
;

require_list:
require_list KW_TYPES
| require_list TK_KEYWORD
| /* empty */
;

// predicate declarations

domain_predicates:
TK_OPEN KW_PREDS predicate_list TK_CLOSE
| TK_OPEN KW_PREDS error TK_CLOSE
{
  log_error( "syntax error in (:predicates ...)" );
  yyerrok;
}
;

predicate_list:
predicate_decl predicate_list
| predicate_decl
;

predicate_decl:
TK_OPEN TK_NEW_SYMBOL
{
  current_param.clear();
}
typed_param_list TK_CLOSE
{
  PredicateSymbol *p = new PredicateSymbol( $2->text );
  dom_predicates.push_back( p );
  p->param = current_param;
  clear_param( current_param );
  $2->val = p;
}
| TK_OPEN error TK_CLOSE
{
  log_error( "syntax error in predicate declaration" );
  yyerrok;
}
;

typed_param_list:
typed_param_list typed_param_sym_list TK_HYPHEN TK_TYPE_SYMBOL
{
  set_variable_type( current_param, current_param.size(), (TypeSymbol*)$4->val );
}
| typed_param_list typed_param_sym_list
{
  set_variable_type( current_param, current_param.size(), dom_top_type );
}
| /* empty */
;

typed_param_sym_list:
typed_param_sym_list TK_NEW_VAR_SYMBOL
{
  $2->val = new VariableSymbol( $2->text );
  current_param.push_back( (VariableSymbol*)$2->val );
}
| /* empty */
;

// type declarations

domain_types:
TK_OPEN KW_TYPES typed_type_list TK_CLOSE
;

typed_type_list:
typed_type_list primitive_type_list TK_HYPHEN TK_TYPE_SYMBOL
{
  set_type_type( dom_types, dom_types.size(), (TypeSymbol*)$4->val );
}
| typed_type_list primitive_type_list TK_HYPHEN TK_NEW_SYMBOL
{
  $4->val = new TypeSymbol( $4->text );
  if( write_warnings ) std::cerr << "warning: assuming " << $4->text << " - object" << std::endl;
  ((TypeSymbol*)$4->val)->sym_type = dom_top_type;
  set_type_type( dom_types, dom_types.size(), (TypeSymbol*)$4->val );
  dom_types.push_back( (TypeSymbol*)$4->val );
}
| typed_type_list primitive_type_list
{
  set_type_type( dom_types, dom_types.size(), dom_top_type );
}
| /* empty */
;

primitive_type_list:
primitive_type_list TK_TYPE_SYMBOL
{
  /* the type is already (implicitly) declared */
}
| primitive_type_list TK_NEW_SYMBOL
{
  $2->val = new TypeSymbol( $2->text );
  dom_types.push_back( (TypeSymbol*)$2->val );
}
| /* empty */
;

// constant declarations

domain_constants:
TK_OPEN KW_CONSTANTS typed_constant_list TK_CLOSE
| TK_OPEN KW_OBJECTS typed_constant_list TK_CLOSE
;

typed_constant_list:
typed_constant_list ne_constant_sym_list TK_HYPHEN TK_TYPE_SYMBOL
{
  set_constant_type( dom_constants, dom_constants.size(), (TypeSymbol*)$4->val );
}
| typed_constant_list ne_constant_sym_list
{
  set_constant_type( dom_constants, dom_constants.size(), dom_top_type );
}
| /* empty */
;

ne_constant_sym_list:
ne_constant_sym_list TK_NEW_SYMBOL
{
  $2->val = new Symbol( $2->text );
  dom_constants.push_back( (Symbol*)$2->val );
}
| TK_NEW_SYMBOL
{
  $1->val = new Symbol( $1->text );
  dom_constants.push_back( (Symbol*)$1->val );
}
;

// structure declarations

domain_structure:
action_declaration
;

// structure declarations

action_declaration:
TK_OPEN KW_ACTION action_symbol
{
  ActionSymbol *a = new ActionSymbol( $3->text );
  dom_actions.push_back( a );
  context.push_front( a );
}
action_elements TK_CLOSE
{
  // post-processing should actually be done on all actions after the complete
  // domain and problem have been read (calling PDDL_Base::post_process())
  dom_actions.back()->post_process();
  clear_param( current_param );
  $3->val = context.front();
  context.pop_front();
}
| TK_OPEN KW_ACTION error TK_CLOSE
{
  log_error( "syntax error in action declaration" );
  yyerrok; 
}
;

action_elements:
action_elements KW_ARGS TK_OPEN
{
  current_param.clear();
}
typed_param_list TK_CLOSE
{
  ((Complex*)context.front())->param = current_param;
}
| action_elements KW_EFFECT action_effect
| action_elements KW_PRE condition
| /* empty */
;

condition:
single_condition
| TK_OPEN KW_AND condition_list TK_CLOSE
;

condition_list:
single_condition condition_list
| single_condition
;

single_condition:
positive_atom
| negative_atom
;

positive_atom:
TK_OPEN TK_PRED_SYMBOL
{
  current_atom = new Atom( (PredicateSymbol*)$2->val );
}
atom_argument_list TK_CLOSE
{
  context.front()->pos_atm.push_back( (Atom*)current_atom );
}
| TK_OPEN TK_EQ arg_symbol arg_symbol TK_CLOSE
{
  Atom* eq_atom = new Atom( dom_eq_pred );
  eq_atom->param.push_back( (Symbol*)$3->val );
  eq_atom->param.push_back( (Symbol*)$4->val );
  context.front()->pos_atm.push_back( eq_atom );
}
;

negative_atom:
TK_OPEN KW_NOT TK_OPEN TK_PRED_SYMBOL
{
  current_atom = new Atom( (PredicateSymbol*)$4->val );
  current_atom->neg = true;
}
atom_argument_list TK_CLOSE TK_CLOSE
{
  context.front()->neg_atm.push_back( (Atom*)current_atom );
}
| TK_OPEN KW_NOT TK_OPEN TK_EQ arg_symbol arg_symbol TK_CLOSE TK_CLOSE
{
  Atom* eq_atom = new Atom(dom_eq_pred);
  eq_atom->neg = true;
  eq_atom->param.push_back( (Symbol*)$5->val );
  eq_atom->param.push_back( (Symbol*)$6->val );
  context.front()->neg_atm.push_back( eq_atom );
}
;

atom_argument_list:
atom_argument_list TK_VAR_SYMBOL
{
  if ($2->val == 0) {
    log_error( "undeclared variable in atom args list" );
  }
  else {
    current_atom->param.push_back( (VariableSymbol*)$2->val );
  }
}
| atom_argument_list TK_OBJ_SYMBOL
{
  current_atom->param.push_back( (Symbol*)$2->val );
}
| /* empty */
;

action_effect:
single_action_effect
| TK_OPEN KW_AND action_effect_list TK_CLOSE
;

action_effect_list:
single_action_effect action_effect_list
| single_action_effect
;

single_action_effect:
TK_OPEN KW_FORALL TK_OPEN
{
  stored_n_param = current_param.size();
  context.push_front( new Complex() );
}
typed_param_list TK_CLOSE
{
  Complex* s = (Complex*)context.front();
  for( size_t k = 0; k < current_param.size(); ++k )
    s->param.push_back( current_param[k] );
}
quantified_effect_body TK_CLOSE
{
  Complex *c = (Complex*)context.front();
  context.pop_front();
  ((ActionSymbol*)context.front())->complex.push_back( c );
  clear_param( current_param, stored_n_param );
  while( current_param.size() > stored_n_param ) current_param.pop_back();
}
| TK_OPEN KW_WHEN
{
  Complex* c = new Complex();
  for( size_t k = 0; k < current_param.size(); ++k )
    c->param.push_back( current_param[k] );
  context.push_front( c );
}
condition atomic_effect_kw_list TK_CLOSE
{
  Complex *c = (Complex*)context.front();
  context.pop_front();
  ((ActionSymbol*)context.front())->complex.push_back( c );
}
| atomic_effect
| cnf_clause
{
  ((Complex*)context.front())->clauses.push_back( $1 );
}
| oneof_effect
{
  ((ActionSymbol*)context.front())->oneof.push_back( $1 );
}
;

quantified_effect_body:
TK_OPEN KW_WHEN condition atomic_effect_kw_list TK_CLOSE
| atomic_effect_kw_list
| clause_kw_list
;

atomic_effect_kw_list:
TK_OPEN KW_AND atomic_effect_list TK_CLOSE
| atomic_effect
;

atomic_effect_list:
atomic_effect_list atomic_effect
| atomic_effect
;

atomic_effect:
positive_atom_effect
| negative_atom_effect
;

positive_atom_effect:
TK_OPEN TK_PRED_SYMBOL
{
  current_atom = new Atom( (PredicateSymbol*)$2->val );
}
atom_argument_list TK_CLOSE
{
  context.front()->adds.push_back( (Atom*)current_atom );
}
;

negative_atom_effect:
TK_OPEN KW_NOT TK_OPEN TK_PRED_SYMBOL
{
  current_atom = new Atom( (PredicateSymbol*)$4->val, true );
}
atom_argument_list TK_CLOSE TK_CLOSE
{
  context.front()->dels.push_back( (Atom*)current_atom );
}
;

cnf_clause:
TK_OPEN KW_OR
{
  context.push_front( new Clause() );
}
condition_list TK_CLOSE
{
  $$ = (Clause*)context.front();
  context.pop_front();
}
;

clause_kw_list:
TK_OPEN KW_AND clause_list TK_CLOSE
| cnf_clause
{
  ((Complex*)context.front())->clauses.push_back( $1 );
}
;

clause_list:
clause_list cnf_clause
{
  ((Complex*)context.front())->clauses.push_back( $2 );
}
| cnf_clause
{
  ((Complex*)context.front())->clauses.push_back( $1 );
}
;

oneof_effect:
TK_OPEN KW_ONEOF
{
  context.push_front( new OneOf() );
}
atomic_effect_list TK_CLOSE
{
  $$ = (OneOf*)context.front();
  context.pop_front();
}
;

// problem definition

pddl_problem:
TK_OPEN KW_DEFINE TK_OPEN KW_PROBLEM any_symbol TK_CLOSE
{
  problem_name = $5->text;
}
problem_elements TK_CLOSE
| TK_OPEN KW_DEFINE TK_OPEN KW_PROBLEM error TK_CLOSE
{
  log_error( "syntax error in problem definition." );
  yyerrok;
}
;

problem_elements:
problem_elements TK_OPEN KW_FORDOMAIN any_symbol TK_CLOSE
| problem_elements domain_requires
| problem_elements domain_constants
| problem_elements initial_state
| problem_elements goal_spec
| /* empty */
;

initial_state:
TK_OPEN KW_INIT TK_OPEN KW_AND init_elements TK_CLOSE TK_CLOSE
| TK_OPEN KW_INIT init_elements TK_CLOSE
;

init_elements:
init_elements init_atom
| init_elements cnf_clause
{
  dom_init_cls.push_back( $2 );
}
| init_elements oneof_effect
{
  dom_init_oneof.push_back( $2 );
}
| /* empty */
;

init_atom:
TK_OPEN TK_PRED_SYMBOL
{
  current_atom = new Atom((PredicateSymbol*)$2->val);
}
atom_argument_list TK_CLOSE
{
  PredicateSymbol* p = (PredicateSymbol*)$2->val;
  if (p->param.size() != current_atom->param.size()) {
    log_error( "wrong number of arguments for predicate in (:init ..." );
  }
  //insert_atom(p->init, (Atom*)current_atom);
  dom_init_atoms.push_back( (Atom*)current_atom );
}
;

goal_spec:
TK_OPEN KW_GOAL single_goal TK_CLOSE
| TK_OPEN KW_GOAL cnf_clause TK_CLOSE
{
  dom_goal_cls.push_back( $3 );
}
| TK_OPEN KW_GOAL TK_OPEN KW_AND goal_list TK_CLOSE TK_CLOSE
;

goal_list:
single_goal goal_list
| cnf_clause goal_list
{
  dom_goal_cls.push_back( $1 );
}
| single_goal
| cnf_clause
{
  dom_goal_cls.push_back( $1 );
}
;

single_goal:
positive_atom_goal
| negative_atom_goal
;

positive_atom_goal:
TK_OPEN TK_PRED_SYMBOL
{
  current_atom = new Atom((PredicateSymbol*)$2->val);
}
atom_argument_list TK_CLOSE
{
  dom_goal_pos.push_back( (Atom*)current_atom );
  insert_atom(((Atom*)current_atom)->pred->pos_goal, (Atom*)current_atom);
}
;

negative_atom_goal:
TK_OPEN KW_NOT TK_OPEN TK_PRED_SYMBOL
{
  current_atom = new Atom((PredicateSymbol*)$4->val);
  current_atom->neg = true;
}
atom_argument_list TK_CLOSE TK_CLOSE
{
  dom_goal_neg.push_back( (Atom*)current_atom );
  insert_atom(((Atom*)current_atom)->pred->neg_goal, (Atom*)current_atom);
}
;

clause_kw_list_goal:
TK_OPEN KW_AND clause_list_goal TK_CLOSE
| cnf_clause
{
  dom_goal_cls.push_back( $1 );
}
;

clause_list_goal:
clause_list_goal cnf_clause
{
  dom_goal_cls.push_back( $2 );
}
| cnf_clause
{
  dom_goal_cls.push_back( $1 );
}
;

%%
