#ifndef STATE_H
#define STATE_H

#include "problem.h"
#include "md4.h"
#include <iostream>
#include <ext/hash_map>
#include <ext/hash_set>
#include <list>
#include <assert.h>

class StateHash;
class StateSet;

class State {
  size_t capacity_;
  size_t size_;
  size_t *atoms_;
  void make_space( size_t n )
  {
    capacity_ = (capacity_?capacity_:2);
    while( capacity_ < 2*n ) capacity_ *= 2;
    size_t *atoms = new size_t[capacity_+1];
    for( size_t *p = atoms, *q = atoms_; *q; ) *p++ = *q++;
    atoms[size_] = 0;
    delete[] atoms_;
    atoms_ = atoms;
  }
public:
  State( size_t n = 0 ) : capacity_(n), size_(0), atoms_(new size_t[capacity_+1]) { atoms_[0] = 0; }
  State( const State &s ) : capacity_(s.size_), size_(s.size_), atoms_(new size_t[capacity_+1]) { for( size_t *p = atoms_, *q = s.atoms_; *q; ) *p++ = *q++; atoms_[size_] = 0; }
  State( const index_set &s ) : capacity_(s.size()), size_(0), atoms_(new size_t[capacity_+1]) { atoms_[0] = 0; add(s); }
  virtual ~State() { delete[] atoms_; }
  void clear() { atoms_[0] = 0; size_ = 0; }
  size_t hash() const
  {
    MD4_CTX md4_ctx;
    unsigned char dig[16];
    MD4Init( &md4_ctx );
    MD4Update( &md4_ctx, (unsigned char*)atoms_, size_ * sizeof(size_t) );
    MD4Final( dig, &md4_ctx );
    size_t *p = (size_t*)dig;
    size_t result = p[0] ^ p[1] ^ p[2] ^ p[3];
    return( result );
  }

  bool satisfy( register int a, bool neg = false ) const
  {
    ++a;
    for( register size_t *p = atoms_; *p; p++ )
      if( *p == a ) return( !neg );
    return( neg );
  }
  bool satisfy( const index_set &pre ) const
  {
    for( index_set::const_iterator pi = pre.begin(); pi != pre.end(); ++pi ) {
      int l = (*pi), p = (l<0?-l:l);
      if( !satisfy(p-1,l!=p) ) return( false );
    }
    return( true );
  }
  bool satisfy( const index_vec &clause ) const
  {
    for( size_t k = 0; k < clause.size(); ++k ) {
      int l = clause[k];
      int p = (l<0?-l:l);
      if( satisfy(p-1,l!=p) ) return( true );
    }
    return( false );
  }
  bool satisfy( const index_vec_vec &cls ) const
  {
    for( size_t k = 0; k < cls.size(); ++k )
      if( !satisfy(cls[k]) ) return( false );
    return( true );
  }

  void add( register size_t a )
  {
    ++a;
    make_space( size_ + 1 );
    register size_t *p = atoms_;
    while( *p && (*p < a) ) ++p;
    if( !*p ) {
      ++size_;
      *p++ = a;
      *p = 0;
    }
    else if( *p != a ) {
      atoms_[size_+1] = 0;
      for( register size_t *q = &atoms_[size_]; q > p; q-- ) *q = *(q-1);
      *p = a;
      ++size_;
    }
  }
  void add( const index_set &s ) { for( index_set::const_iterator si = s.begin(); si != s.end(); ++si ) add(*si); }
  void remove( register size_t a )
  {
    ++a;
    register size_t *p;
    for( p = atoms_; *p && (*p < a); ++p );
    if( *p == a ) {
      do {
        *p = *(p+1);
        ++p;
      } while( *p );
      --size_;
    }
  }
  void remove( const index_set &s ) { for( index_set::const_iterator si = s.begin(); si != s.end(); ++si ) remove(*si); }

  void apply( int p, bool neg ) { if( neg ) remove(p); else add(p); }
  void apply( const index_set &a, const index_set &d ) { remove(d); add(a); }
  void conditional_apply( const State &s, const index_set &pre, const index_set &add, const index_set &del ) { if( s.satisfy(pre) ) apply( add, del ); }
  void apply( const Instance &ins, size_t k ) { apply( ins.actions[k]->add, ins.actions[k]->del ); }
  void apply( const Instance::Action &a, StateSet &bel ) const;

  bool goal( const Instance &ins ) const
  {
    for( index_set::const_iterator pi = ins.goal_atoms.begin(); pi != ins.goal_atoms.end(); ++pi )
      if( !satisfy(*pi) ) return( false );
    return( satisfy(ins.goal_cls) );
  }
  bool applicable( const Instance &ins, size_t k ) const { return( satisfy(ins.actions[k]->pre) ); }

  bool operator==( const State &s ) const { if( size_ != s.size_ ) return( false ); size_t *p = atoms_, *q = s.atoms_; while( *p && (*p++ == *q++) ); return( !*p ); }
  bool operator<( const State &s ) const { if( size_ != s.size_ ) return( size_ < s.size_ ); for( size_t *p = atoms_, *q = s.atoms_; *p; ) if( *p++ != *q++ ) return( *(p-1) < *(q-1) ); return( false ); }
  const State& operator=( const State &s ) { make_space( s.size_ ); size_ = s.size_; for( size_t *p = atoms_, *q = s.atoms_; *q; ) *p++ = *q++; atoms_[size_] = 0; }
  bool operator()( const State *s1, const State *s2 ) const { return( *s1 == *s2 ); }
  size_t operator()( const State *s ) const { return( s->hash() ); }

  void print( std::ostream &os, const Instance &i ) const { os << "{"; for( size_t *p = atoms_; *p; ++p ) os << i.atoms[(*p)-1]->name << (*(p+1)?",":""); os << "}"; }
  void write( std::ostream &os ) const { os << size_ << ":{"; for( size_t *p = atoms_; *p; ++p ) os << (*p)-1 << (*(p+1)?",":""); os << "}"; }

  struct ptr_map_fun {
    bool operator()( const State *p1, const State *p2 ) const { return( (*p1) < (*p2) ); }
  };
};

inline std::ostream& operator<<( std::ostream &os, const State &s )
{
  s.write( os );
  return( os );
}

namespace Hash {

  template<typename T> class hash {
  public:
    size_t operator()( const T &s ) const { return( s.hash() ); }
  };

  template<typename T> class equal_to {
  public:
    size_t operator()( const T &s1, const T &s2 ) const { return( s1 == s2 ); }
  };

  class Data {
    size_t action_;
    mutable size_t marks_;
    mutable size_t scc_low_;
    mutable size_t scc_idx_;
  public:
    Data( size_t action = UINT_MAX, size_t marks = 0 ) : action_(action), marks_(marks), scc_low_(UINT_MAX), scc_idx_(UINT_MAX) { }
    size_t action() const { return( action_ ); }
    void set_action( size_t action ) { action_ = action; }
    bool marked() const { return( marks_&0x1 ); }
    void mark() { marks_ |= 0x1; }
    bool solved() const { return( marks_&0x2 ); }
    void solve() { marks_ |= 0x2; }
    size_t scc_low() const { return( scc_low_ ); }
    void set_scc_low( size_t low ) const { scc_low_ = low; }
    size_t scc_idx() const { return( scc_idx_ ); }
    void set_scc_idx( size_t idx ) const { scc_idx_ = idx; }
    void write( std::ostream &os ) const { os << "{" << action_ << "," << marks_ << "}"; }
  };

};

inline std::ostream& operator<<( std::ostream &os, const Hash::Data &d )
{
  d.write( os );
  return( os );
}

namespace hashing = ::__gnu_cxx;
namespace __gnu_cxx {

  template<typename T, typename F=Hash::hash<T>, typename E=Hash::equal_to<T> > class Hash_set : public hash_set<T,F,E> { };
  template<typename T, typename F=Hash::hash<T>, typename E=Hash::equal_to<T>  > class Hash_map : public hash_map<T,Hash::Data*,F,E> {
    typedef hash_map<T,Hash::Data*,F,E> base_;
  public:
    typedef typename hash_map<T,Hash::Data*,F,E>::iterator iterator;
    typedef typename hash_map<T,Hash::Data*,F,E>::const_iterator const_iterator;
    const_iterator begin() const { return( base_::begin() ); }
    const_iterator end() const { return( base_::end() ); }
    iterator begin() { return( base_::begin() ); }
    iterator end() { return( base_::end() ); }
  protected:
    Hash::Data* push( const T &s, Hash::Data *d ) { std::pair<iterator,bool> p = insert( std::make_pair(s,d) ); return( d ); }
    iterator lookup( const T &s ) { return( find(s) ); }
    const_iterator lookup( const T &s ) const { return( find(s) ); }
  public:
    Hash_map() { }
    virtual ~Hash_map() { for( iterator hi = begin(); hi != end(); ++hi ) delete (*hi).second; }
    Hash::Data* data_ptr( const T &s ) { iterator di = lookup(s); if( di == end() ) return( push(s,new Hash::Data) ); else return( (*di).second ); }
    bool marked( const T &s ) const { const_iterator di = lookup(s); return( di == end() ? false : (*di).second->marked() ); }
    void mark( const T &s ) { iterator di = lookup(s); if( di == end() ) push(s,new Hash::Data(UINT_MAX,1)); else (*di).second->mark(); }
    bool solved( const T &s ) const { const_iterator di = lookup(s); return( di == end() ? false : (*di).second->solved() ); }
    void solve( const T &s ) { iterator di = lookup(s); if( di == end() ) push(s,new Hash::Data(UINT_MAX,2)); else (*di).second->solve(); }
    size_t action( const T &s ) const { const_iterator di = lookup(s); if( di == end() ) return( UINT_MAX ); else return( (*di).second->action() ); }
    void set_action( const T &s, size_t action ) { iterator di = lookup(s); if( di == end() ) push(s,new Hash::Data(action,false)); else (*di).second->set_action( action ); }
    void dump( std::ostream &os ) const { for( const_iterator di = begin(); di != end(); ++di ) os << (*di).first << " : " << *(*di).second << std::endl; }
  };

};

class StateHash : public hashing::Hash_map<const State*,State,State> { };
class StateSet : public hashing::Hash_set<const State*,State,State> { };

inline void State::apply( const Instance::Action &a, StateSet &bel ) const
{
  static State s, t;

  s = *this;
  s.apply( a.add, a.del );
  for( size_t k = 0; k < a.when.size(); ++k )
    s.conditional_apply( *this, a.when[k].pre, a.when[k].add, a.when[k].del );
  t = s;

  int i = 0, j = 0;
  std::vector<int> stack, restore;
  while( true ) {
    assert( stack.size() == restore.size() );
    if( i == a.oneof.size() ) {
      if( t.satisfy(a.cls) && (bel.find(&t) == bel.end()) ) {
        State *x = new State(t);
        bel.insert( x );
      }
      if( i == 0 ) break;
      int l = restore.back(), p = (l<0?-l:l);
      j = stack.back();
      if( l != INT_MAX ) t.apply( p-1, p!=l );
      restore.pop_back();
      stack.pop_back();
      --i;
      continue;
    }
    if( j == a.oneof[i].size() ) {
      if( i == 0 ) break;
      int l = restore.back(), p = (l<0?-l:l);
      j = stack.back();
      if( l != INT_MAX ) t.apply( p-1, p!=l );
      restore.pop_back();
      stack.pop_back();
      --i;
      continue;
    }
    int l = a.oneof[i][j], p = (l<0?-l:l);
    restore.push_back( t.satisfy(p-1,p==l) ? -l : INT_MAX );
    t.apply( p-1, p!=l );
    stack.push_back( 1+j );
    ++i;
    j = 0;
  }
}

#endif

